@@ -807,10 +807,6 @@ static PyObject* Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kw
             exit(-1);
         }
 
-        // XXX: We need to wait for the server to spawn.  There's probably a
-        // better way.
-        usleep(100000);
-
         moduleConfiguration.socket = (struct sockaddr_un) { .sun_family = AF_UNIX };
         snprintf((char*) &moduleConfiguration.socket.sun_path,
                 sizeof(moduleConfiguration.socket.sun_path),
@@ -823,10 +819,23 @@ static PyObject* Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kw
             exit(-1);
         }
 
-        if (connect(sock, (struct sockaddr*) &moduleConfiguration.socket,
-                    sizeof(moduleConfiguration.socket)) == -1) {
-            perror("Failed to connect to TraceInitialization socket");
-            exit(-1);
+        // Wait for 1s for the functiontrace-server to be ready to receive
+        // connections.
+        for (size_t tries = 0; true; tries++) {
+            if (connect(sock, (struct sockaddr*) &moduleConfiguration.socket,
+                        sizeof(moduleConfiguration.socket)) == 0) {
+                break;
+            }
+
+            if (tries == 1000) {
+                // If we haven't connected by now, we never will.
+                perror("Failed to connect to TraceInitialization socket");
+                exit(-1);
+            }
+
+            // We sleep 1ms at a time since we're profiling python and can
+            // afford to wait.
+            usleep(1000);
         }
 
         // Register an env var so subprocesses know how to connect to our
@@ -997,6 +1006,15 @@ static ThreadState* Fprofile_CreateThreadState(void) {
     }
     mpack_finish_map(&state->writer);
 
+    // Flush the state of our new thread to ensure that we'll record
+    // *something* for it in the case where we quickly exit after opening the
+    // socket.
+    //
+    // TODO: We should have a different approach that doesn't require explicit
+    // flushing.  The obvious candidate is shared-mem, but that opens us up to
+    // a whole pile of issues we'd hoped to avoid.
+    mpack_writer_flush_message(&state->writer);
+
     return state;
 }
 
@@ -1023,17 +1041,6 @@ static void Fprofile_ResetThreadState(void) {
     // already share our moduleConfiguration.  However, we do need to record
     // the new pid since we're in a new process.
     moduleConfiguration.pid = getpid();
-    ThreadState* newState = Fprofile_CreateThreadState();
-
-    // Flush the state of our new thread to ensure that we'll record
-    // *something* for it in the case where we quickly exit after opening the
-    // socket.
-    // This shouldn't happen (we should catch exit cases and force a flush
-    // then), but it can't hurt to be safe.
-    //
-    // TODO: We should have a different approach that doesn't require explicit
-    // flushing.  The obvious candidate is shared-mem, but that opens us up to
-    // a whole pile of issues we'd hoped to avoid.
-    mpack_writer_flush_message(&newState->writer);
+    Fprofile_CreateThreadState();
 }
 
